/*
 * Bend Matrix 2
 * 
 * ATMEGA32
 * 8MHz --> 16MHz
 *
 
 2x4x4 buttons is one matrix or patch = 32 bits
 8 matrices per patches is one sequence or bank= 256 bits
 16 sequences or banks = 4096 bits or 512 Bytes (ATMEGA32 has 1024 Bytes of EEPROM)
*/


#include <stdarg.h>
#include <ctype.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>

#include "bendmatrix2.h"
#include "effects.h"
#include "inputs.h"
#include "memory.h"
#include "timer.h"
#include "midi.h"
#include "outputs.h"
#include "uart.h"

uint8_t ain[4]={0,0,0,0};				//analog inputs (1-4)

uint16_t matrix[2];						//matrix[color]: matrix[0] is red, matrix[1] is green
uint16_t midimatrix[2];
uint16_t buttonmatrix[2];
uint16_t s_matrix[16][8][2];			//s_matrix[bank][step][color] 16 banks, each a sequence of 8 steps of 32 bits each
uint16_t switch_matrix[2];
uint16_t latchoff_matrix[2];
uint16_t latchon_matrix[2];

#ifdef GREEN_IS_BEATS
	uint16_t green_beat_matrix=0;
	uint16_t green_beat_clear_matrix=0;
#endif

uint8_t curbank=0;
uint8_t seq_leds=1;					//state of seqeuncer leds
uint8_t cur;								//current matrix/patch/"step" in the sequencer

char timerflash=0;							//flag used to flash the bpm light
int play_timer=0,flash_timer=0; 			//variables used by the timer routines
int bank_indicate_flash=0;
volatile unsigned int newbpm;					//the rotary encoder updated by interrupts
unsigned int bpm;

/* eeprom stored globals:*/
uint8_t knob_mode=0,sub_knob_mode=0,path_mode=0,knob_layout_mode=1,midi_mode=3;

uint8_t save_mode=0,play_mode=0;		//indicates our mode
uint8_t bank_mode=0;
uint8_t save_bank_mode=0,length_mode=0;

uint8_t just_stepped=0;
short clockreceived=0;
short ainchanged=0;

uint8_t	hold_menu_display=0;
uint8_t assign_midi_mode=0,assign_midi_curbut=0;
volatile uint8_t new_assign_midi_value;				//the rotary encoder updated by interrupts
uint8_t assign_midi_value;
uint8_t assigned_midi_notes[32];
#ifdef MAP_CCS_TO_BUTTONS
	uint8_t assigned_midi_CCbuttons[32];
#endif
uint8_t CC_number[4];
uint8_t show_midi_menu_display;

unsigned short seven_seg_key;		//for indicating on the display
char seven_seg[4];
uint8_t bpmchange=0;

uint8_t patch_play_start[16], patch_play_end[16];
uint8_t bank_play_start, bank_play_end;

uint8_t trig_playpause=0;
uint8_t onoffcycle_mode=1;
long onoffcycle_timer=0;
long onoffcycle_lastevent=0;

uint8_t KNOBMODENAMES[4]={'S'-ALPHA,'A'-ALPHA,'G'-ALPHA,'R'-ALPHA};

char* LENGTHNAMES[5]={"BEG","END","END","END","END"};
char* REPEATNAMES[4]={"FWD", "REV", "RAN", "ONE"};
char* KLNAMES[3]={"^ O", "___", "|||"};

char* MIDIMODENAMES[8]={"SND", "RCV", "ALL", "OFF", "NOC", "CLK", "HLF", "DBL"};



uint8_t DUALMATRIXMODE;



inline short BIT(short bit){	return (1<<bit);}


ISR(INT0_vect){
//TEMPO UP
//Interrupt is called when INT0 falls from 1 to 0
//So if INT1 is 1, then the greycode went from 11 to 01
	if (TEMPO & BIT(DOWN)) 
	{
		if (!assign_midi_mode){
			if (newbpm<2000) {
				if (SHIFT) newbpm+=5;
				else if (newbpm<100) newbpm+=10;
				else newbpm+=50;
			}
			bpmchange=1;
		}

		else if (new_assign_midi_value<127)	new_assign_midi_value++;
	
		seven_seg_key=1;
	}
}

ISR(INT1_vect){
//Interrupt is called when INT1 falls from 1 to 0
//So if INT0 is 1, then the greycode went from 11 to 10
	if (TEMPO & BIT(UP)) {
		if (!assign_midi_mode){
			 if(newbpm>0) {
				if (SHIFT) 	newbpm-=5;
				else if (newbpm>100) newbpm-=50;
				else if (newbpm>10) newbpm-=10;
				else newbpm=0;
			}
			bpmchange=1;
		}

		else if (new_assign_midi_value>0) new_assign_midi_value--;

		seven_seg_key=1;

	}
}

void init_rotary(void){
	MCUCR &= ~(BIT(0) | BIT(2)); 
	MCUCR |= (BIT(1) | BIT(3));     //ISC11=1, ISC10=0, ISC01=1, ISC00=0
							        //which means falling edge on INT1 or INT0 generates interrupt

	GICR |= BIT(INT1) | BIT(INT0);	//enable INT0 and INT1 as interrupts
}

void initports(void){
	
	SR_ddr |= BIT(SR_CP) | BIT(SR_PL);

	OSR_ddr|= BIT(OSR_ST) | BIT(OSR_SH) | BIT(OSR_DS);

	OSR |= BIT(OSR_SH);
	OSR |= BIT(OSR_DS);
	OSR |= BIT(OSR_ST);

	//init the ADC:
	DDRA &= ~(BIT(AIN0) | BIT(AIN1) | BIT(AIN2) | BIT(AIN3));			//ADC pins 0,1 to inputs
	ADCSRA = BIT(ADEN);	//Enable On
	ADMUX = BIT(ADLAR);	//Left-Adjust


		//TEMPO UP/DOWN BUTTONS inputs, pull-down
	TEMPO_ddr &= ~BIT(UP);
	TEMPO_ddr &= ~BIT(DOWN);
	TEMPO &= ~BIT(UP);
	TEMPO &= ~BIT(DOWN);

	DDRA &= BIT(2);	//MOMENTMODE is input
	PORTA |= BIT(2); //pull-up


	PLAYSAVE_ddr |= BIT(PLAY_LED) | BIT(SAVE_LED); //LED outputs
	PLAYSAVE_ddr &= ~PLAYSAVECLEARBUTTONS; //button inputs
	PLAYSAVE_LED &= ~(BIT(PLAY_LED) | BIT(SAVE_LED));



	//MM8816 init: set address/data/control pins to output(DDR), and set to zero
	SWITCH_ddr=0xFF;
	SWITCH = 0x00;

	SWITCH_HI_ddr |= BIT(SW_AX2) | BIT(SW_AX3);
	SWITCH_HI &= ~(BIT(SW_AX2) | BIT(SW_AX3));

	SWITCH |= BIT(SW_RESET);
	SWITCH &= ~BIT(SW_RESET);



	//tempo LED
	TEMPO_ddr |= BIT(TEMPO_LED);
	TEMPO_LED_PORT &= ~BIT(TEMPO_LED);

	EECR &= ~(1<<EERIE);
}

uint16_t get_use_bpm(void){
//	return ((bpm[curbank]==0)?2:bpm[curbank]);
	return ((bpm==0)?2:bpm);
}

void get_nextstep(uint8_t *nextbank, uint8_t *nextcur){

	if(cur!=patch_play_end[curbank]){ //if we're not at the end of the current bank...
		*nextbank=curbank;				//then we just go to the next patch
		*nextcur=cur+1;
	} else {
		if (!bank_mode) {
			*nextcur=patch_play_start[curbank];
			*nextbank=curbank;
		} else {
			if (curbank==bank_play_end) *nextbank=bank_play_start;
			else *nextbank=curbank+1;
			*nextcur=patch_play_start[*nextbank];
		}
	}

}

void get_prevstep(uint8_t *prevbank, uint8_t *prevcur){

	if(cur!=patch_play_start[curbank]){ //if we're not at the end of the current bank...
		*prevbank=curbank;				//then we just go to the next patch
		*prevcur=cur-1;
	} else {
		if (!bank_mode) {
			*prevcur=patch_play_end[curbank];
			*prevbank=curbank;
		} else {
			if (curbank==bank_play_start) *prevbank=bank_play_end;
			else *prevbank=curbank-1;
			*prevcur=patch_play_end[*prevbank];
		}
	}

}


void do_play_mode(void){

	static uint8_t SPINt=0;
	static int midiclock_timer=0;
		#ifdef GREEN_IS_BEATS
		static int greenbeats_timer=0;
		static short green_beats_flash=0;
		short i;
		#endif
	short use_bpm;
	uint8_t nextbank, nextcur;	

	use_bpm=get_use_bpm();

	#ifdef GREEN_IS_BEATS
	if (DUALMATRIXMODE==2){
		if (inttimerreached(&greenbeats_timer, (15000/use_bpm)*2)) { 
			if (++green_beats_flash>=4) green_beats_flash=0;

			uint16_t mm=buttonmatrix[1] | matrix[1];

			for (i=0;i<4;i++){ 

				green_beat_matrix &= ~BIT(i*4+2);
				green_beat_clear_matrix &= ~BIT(i*4+2);

				if (
				   ((mm & BIT(i*4+0)) && (green_beats_flash==0))
				|| ((mm & BIT(i*4+1)) && (green_beats_flash==2))
				|| ((mm & BIT(i*4+2)) && (green_beats_flash==0))
				|| ((mm & BIT(i*4+2)) && (green_beats_flash==2))
				|| ((mm & BIT(i*4+3)) && (green_beats_flash==1))
				|| ((mm & BIT(i*4+3)) && (green_beats_flash==3))
				){
					green_beat_matrix |= BIT(i*4+2);
				}

				if (
				   ((mm & BIT(i*4+0)) && (green_beats_flash==2))
				|| ((mm & BIT(i*4+1)) && (green_beats_flash==0))
				|| ((mm & BIT(i*4+2)) && (green_beats_flash==1))
				|| ((mm & BIT(i*4+2)) && (green_beats_flash==3))
				|| ((mm & BIT(i*4+3)) && (green_beats_flash==0))
				|| ((mm & BIT(i*4+3)) && (green_beats_flash==2))
				){
					green_beat_clear_matrix |= BIT(i*4+2);
				}


			}

		}
	}
	#endif	


	if (play_mode || ALWAYS_FLASH_TEMPO){
		if (inttimerreached(&flash_timer, (30000/use_bpm)*2)) { 
			timerflash=!timerflash;
			if (timerflash) TEMPO_LED_PORT |= BIT(TEMPO_LED);
			else TEMPO_LED_PORT &= ~BIT(TEMPO_LED);

			if (play_mode){
				SPINt++;
				if (SPINt>7) SPINt=0;
			}
		}
	} else {
		TEMPO_LED_PORT &= ~BIT(TEMPO_LED);
		flash_timer=0;
	}

	if (play_mode){
		if (midi_mode!=4){ //NO Clock mode

			int midiclocktimer=2500;
			if (midi_mode==6) midiclocktimer=5000;
			if (midi_mode==7) midiclocktimer=1250;

			if (inttimerreached(&midiclock_timer,midiclocktimer/use_bpm)) 
				send_one_MIDI_byte(BPM_CLOCK_MIDI);
		}

		if ((inttimerreached(&play_timer,60000/use_bpm)) || (clockreceived)){
			clockreceived=0;
			play_timer=0;

			if (path_mode==0){ //FWD
				get_nextstep(&nextbank,&nextcur);
				curbank=nextbank;
				cur=nextcur;
			}
			else if (path_mode==3){ //ONE
				get_nextstep(&nextbank,&nextcur);
				if (nextcur==patch_play_start[nextbank]){
					if (!bank_mode) play_mode=0;
					else if (nextbank==bank_play_start) play_mode=0;
				}
				if (play_mode) {
					curbank=nextbank;
					cur=nextcur;
				}
			}
			else if (path_mode==2){ //RAN
				if (bank_mode) curbank=do_random() % (bank_play_end-bank_play_start+1)+ bank_play_start;
				cur=do_random() % (patch_play_end[curbank]-patch_play_start[curbank]+1) + patch_play_start[curbank];
			}
			else if (path_mode==1){ //REV
				get_prevstep(&nextbank,&nextcur);
				curbank=nextbank;
				cur=nextcur;
			}
			if (!hold_menu_display) number_to_string(curbank+1,seven_seg);

			just_stepped=1;

			seq_leds=BIT(cur);
		
		}

		if (!seven_seg_key) {
			seven_seg[0]=SPIN0+SPINt;
			seven_seg_key=DISPLAYHOLD-1;
			//seven_seg_key=0;
		}

	} else {
		PLAYSAVE_LED &= ~BIT(PLAY_LED);
		play_timer=0;
	}

}




/***********

	main()

***********/
int main(void) 
{ 
	uint16_t output_matrix[2]={0,0};
//	uint16_t nullmatrix[2]={0,0};
	short k;
	int itmrms,shift_display_hold=0;
	uint8_t f_matrix[4];
	uint16_t ff_matrix[2];
	short change_dual_matrix_mode_down=0;

	/***************
	** INITIALIZE **
	****************/
	
	initports();
	init_uart(MIDIUBBR);
	inittimer();
	init_rotary();
	read_globals();
	read_midi_assigns();

	for (k=0;k<32;k++) {
		#ifdef MAP_CCS_TO_BUTTONS
			if (assigned_midi_CCbuttons[k]>127) assigned_midi_CCbuttons[k]=k*4;
		#endif

		if (assigned_midi_notes[k]>127) assigned_midi_notes[k]=k*4+1;
		if ((k<4) && (CC_number[k]>127)) CC_number[k]=20+k;
	}
	cur=0;
	curbank=0;

//	read_bpms();
//	newbpm=bpm[curbank];
	bpm=120;
	newbpm=bpm;
	
	read_lengths();

	sei();

	save_mode=0;
	save_bank_mode=0;
	play_mode=0;
	length_mode=0;
	midimatrix[0]=0;
	midimatrix[1]=0;	
	buttonmatrix[0]=0;
	buttonmatrix[1]=0;
	switch_matrix[0]=0;
	switch_matrix[1]=0;
	latchon_matrix[0]=0;
	latchon_matrix[1]=0;
	latchoff_matrix[0]=0;
	latchoff_matrix[1]=0;
	clockreceived=0;
	ainchanged=0;

	seven_seg[0]='B'-ALPHA;
	seven_seg[1]='N'-ALPHA;
	seven_seg[2]='D'-ALPHA;
	seven_seg[3]=0;

	output_shiftreg(0xFF, output_matrix, seven_seg);

	
	matrix[0]=s_matrix[0][0][0];	
	matrix[1]=s_matrix[0][0][1];

	output_shiftreg(0xFF, output_matrix, seven_seg);
	
	#if (MATRIX_VERSION==JOYTOY_ghost)
	DUALMATRIXMODE=1;
	#else
	DUALMATRIXMODE=read_dualmatrix_mode();


	if (READDUALMATRIXMODE){
		#ifdef GREEN_IS_BEATS 
			DUALMATRIXMODE++;
			if (DUALMATRIXMODE>=3) DUALMATRIXMODE=0;
		#else
			DUALMATRIXMODE=1-DUALMATRIXMODE;
		#endif
		save_dualmatrix_mode();
		seven_seg_key=1;
	}
	
	#endif

	if (DUALMATRIXMODE==1){
		seven_seg[0]=2*MZ;
		seven_seg[1]=DASH;
		seven_seg[2]=4;
		seven_seg[3]=0;
	} else if (DUALMATRIXMODE==0) {
		seven_seg[0]=2*MZ;
		seven_seg[1]=DASH;
		seven_seg[2]=8;
		seven_seg[3]=0;
	}
	else if (DUALMATRIXMODE==2){
		seven_seg[0]=4;
		seven_seg[1]=DASH;
		seven_seg[2]='D'-ALPHA;
		seven_seg[3]=0;
	}

/***************
** MAIN LOOP  **
****************/

	for (;;)
	{
		if (SHIFT && PLAY && SAVE){
			if (change_dual_matrix_mode_down==0){
				change_dual_matrix_mode_down=1;
				#ifdef GREEN_IS_BEATS 
					DUALMATRIXMODE++;
					if (DUALMATRIXMODE>=3) DUALMATRIXMODE=0;
				#else
					DUALMATRIXMODE=1-DUALMATRIXMODE;
				#endif
				save_dualmatrix_mode();
				seven_seg_key=1;
				if (DUALMATRIXMODE==1){
					seven_seg[0]=2*MZ;
					seven_seg[1]=DASH;
					seven_seg[2]=4;
					seven_seg[3]=0;
				} else if (DUALMATRIXMODE==0) {
					seven_seg[0]=2*MZ;
					seven_seg[1]=DASH;
					seven_seg[2]=8;
					seven_seg[3]=0;
				}
				else if (DUALMATRIXMODE==2){
					seven_seg[0]=4;
					seven_seg[1]=DASH;
					seven_seg[2]='D'-ALPHA;
					seven_seg[3]=0;
				}
			}

			play_mode=0;save_mode=0;

		}  else change_dual_matrix_mode_down=0;

		if (just_stepped==1) just_stepped=2;
		else if (just_stepped==2) just_stepped=0;


		check_play();
		check_save();

		read_adc();

		if (assign_midi_mode) midi_setup_menu();
		else show_midi_menu_display=0;

/******** DISPLAY ********/
		if (!hold_menu_display){
		
			if (SHIFT && (!seven_seg_key || (seven_seg_key==DISPLAYHOLD-1)/*play mode*/)) {

				/* Holding down SHIFT displays the current settings*/
				if (shift_display_hold++>DISPLAYHOLD){
					number_to_string(get_use_bpm(),seven_seg);
				}
				else{
					if (sub_knob_mode<4) {
						seven_seg[0]=KNOBMODENAMES[knob_mode];
						seven_seg[1]=BLANK;
						seven_seg[2]=sub_knob_mode+1;
					} else {
						seven_seg[0]=DASH;
						seven_seg[1]=DASH;
						seven_seg[2]=DASH;}
				}

			}
			else {

				/* Default display is current bank */
				if (seven_seg_key++>DISPLAYHOLD){
					seven_seg_key=0;
					shift_display_hold=0;
					number_to_string(curbank+1,seven_seg);
					seven_seg[0]=TOPO;
				}
			}
		}


/******** Rotary knob ********/

		if (bpmchange){
			if (newbpm>1995) newbpm=1995;
			bpmchange=0;
			itmrms=gettmrms();

	cli();
			play_timer=itmrms+(bpm*(play_timer-itmrms)/newbpm);
			if (play_timer<0) play_timer=0;
			flash_timer=itmrms+(bpm*(flash_timer-itmrms)/newbpm);
			bpm=newbpm;
	sei();

			number_to_string(get_use_bpm(),seven_seg);
		}


/****PLAY MODE****/

		do_play_mode();

		apply_knob_mode();

		check_clear();


/****READ BUTTONS****/
		load_input_shiftreg();

		seq_leds = read_seqs(seq_leds);
		
		read_buttonmatrix();

		#ifdef TRIGGERSENABLED
			read_trigs();
		#endif
		
		read_midi();



		#ifdef ON_OFF_CYCLE_ENABLED

			onoffcycle_timer=gettmrms();

			if (onoffcycle_mode==1){ //ON
				if (onoffcycle_timer>=(onoffcycle_lastevent+ON_CYCLE_TIME)){
					onoffcycle_mode=0; //turn it off
					onoffcycle_lastevent=onoffcycle_timer;
				}
			} else { //OFF
				if (onoffcycle_timer>=(onoffcycle_lastevent+OFF_CYCLE_TIME)){
					onoffcycle_mode=1; //turn it on
					onoffcycle_lastevent=onoffcycle_timer;
				}
			}

		#endif



/****OUTPUT SWITCHES****/
		for (k=0;k<2;k++)
			switch_matrix[k]=(
				(matrix[k] | midimatrix[k] | latchon_matrix[k]) 
				& ~(latchoff_matrix[k])
				) | buttonmatrix[k];

		#ifdef GREEN_IS_BEATS
			switch_matrix[0] |= green_beat_matrix;
			switch_matrix[0] &= ~green_beat_clear_matrix;
		#endif

		#ifdef ON_OFF_CYCLE_ENABLED
			if (!onoffcycle_mode){ //we should turn all buttons off
				switch_matrix[0]=0;
				switch_matrix[1]=0;
			}
		#endif

		set_switches(switch_matrix);

		output_matrix[0]=switch_matrix[0];
		output_matrix[1]=switch_matrix[1];


/********
	Flash matrix/seq LEDs as needed
********/


		if (save_bank_mode){
			flash_matrix(output_matrix);
		} 

		if (SHIFT) {
			seq_leds=BIT(knob_mode);
			flash_one_bank(output_matrix,curbank);
		} else if (CLEAR){
			flash_matrix_on_matrix(output_matrix,latchoff_matrix,1);
		} else if (SAVE){
			flash_matrix_off_matrix(output_matrix,latchon_matrix,1);
		}

		if (save_mode){
			flash_seqs(cur);
			flash_save();
		} else 
			seq_leds=BIT(cur);
		
		if (length_mode && !save_mode){
			if ((length_mode>0) && (length_mode<=3)){
				flash_seqs(99);
				for (k=0;k<patch_play_start[curbank];k++) seq_leds &= ~BIT(k);
				for (k=patch_play_end[curbank]+1;k<8;k++) seq_leds &= ~BIT(k);
			}
			if ((length_mode==1) || (length_mode>=4)){
				flash_matrix(output_matrix);
				if (MZ==1){
					for (k=0;k<bank_play_start;k++) output_matrix[(k>7)?1:0] &= ~BIT(k & 7);
			 		for (k=bank_play_end+1;k<16;k++) output_matrix[(k>7)?1:0] &= ~BIT(k & 7);
				}else{
					output_matrix[1]=0;
					for (k=0;k<bank_play_start;k++) output_matrix[0] &= ~BIT(k);
			 		for (k=bank_play_end+1;k<16;k++) output_matrix[0] &= ~BIT(k);
				}
			}

			string_to_string(LENGTHNAMES[length_mode-1], seven_seg);
		}

		if (assign_midi_mode==1 || assign_midi_mode==3){
			output_matrix[0]=0;
			output_matrix[1]=0;
			flash_one_bank(output_matrix, assign_midi_curbut);
		}
		if (assign_midi_mode==2){
			output_matrix[0]=0;
			output_matrix[1]=0;
			f_matrix[0]=0;	f_matrix[1]=0;	f_matrix[2]=0;	f_matrix[3]=0;
			f_matrix[assign_midi_curbut]=0xFF;

			switch(knob_layout_mode){
				case(1):
					ff_matrix[0]=rows2squares(f_matrix,0);
					ff_matrix[1]=rows2squares(f_matrix,1);
					break;
				case(0):	
					ff_matrix[0]=f_matrix[0] | (f_matrix[1]<<8);
					ff_matrix[1]=f_matrix[2] | (f_matrix[3]<<8);
					break;
				case(2):
					ff_matrix[0]=cols2squares(f_matrix,0);
					ff_matrix[1]=cols2squares(f_matrix,1);
					break;
			}

			flash_matrix_on_matrix(output_matrix,ff_matrix,2);
		}



/********
	OUTPUT LEDS
********/
		if (bank_mode) seven_seg[3]|=BIT(2);
		else seven_seg[3]&=~BIT(2);
		

		output_shiftreg(seq_leds, output_matrix, seven_seg);
		
	}
} //main()

